import os
import os.path
from tempfile import NamedTemporaryFile
from raft.tasks import task


@task
def ecdsa_key_pair(ctx, name, directory='.', curve='secp256k1'):
    """
    creates an ecdsa key pair
    :param ctx:
    :param name: the name of the vert
    :param str curve: pick one of the available values from
        `openssl ecparam -list_curves`
    :param str directory: the directory in which the cert and key
        will be stored
    """
    key_path = os.path.join(directory, f'{name}.key')
    cert_path = os.path.join(directory, f'{name}.crt')
    ctx.run(
        f'openssl ecparam -name {curve} -genkey -noout '
        f' -out {key_path}')
    command = (
        f'openssl ec -in {key_path} -pubout >{cert_path}'
    )
    ctx.run(command)


@task
def ecdsa_self_signed_cert(ctx, name, directory='.', curve='secp256k1', days=3650):
    """
    creates an ecdsa key pair
    :param ctx:
    :param name: the name of the vert
    :param str curve: pick one of the available values from
        `openssl ecparam -list_curves`
    :param str directory: the directory in which the cert and key
        will be stored
    """
    key_path = os.path.join(directory, f'{name}.key')
    cert_path = os.path.join(directory, f'{name}.crt')
    ctx.run(
        f'openssl ecparam -name {curve} -genkey -noout '
        f' -out {key_path}')
    with NamedTemporaryFile('w') as f:
        f.write('\n'.join([
            '[req]',
            'distinguished_name = req_distinguished_name',
            'req_extensions = v3_req',
            'prompt = no',
            '[req_distinguished_name]',
            'C = US',
            f'CN = {name}',
            '[v3_req]',
            'keyUsage = keyEncipherment, dataEncipherment',
            'extendedKeyUsage = serverAuth, clientAuth',
        ]))
        f.flush()
        command = (
            f'openssl req -new -x509 -sha256'
            f' -out {cert_path}'
            f' -key {key_path}'
            f' -days {days}'
            f' -config {f.name}'
        )
        ctx.run(command)
        # command = (
        #     f'openssl x509 -req -days {days}'
        #     f' -in {csr_path} -signkey {key_path} -out {cert_path}'
        # )
        # ctx.run(command)



from raft import Collection
from .validator import validate_chain
from .ed25519 import ed25519_self_signed_cert
from .ecdsa import ecdsa_key_pair, ecdsa_self_signed_cert


cert_tasks = Collection(
    validate_chain,
    ed25519_self_signed_cert,
    ecdsa_key_pair,
    ecdsa_self_signed_cert
)



import asyncio

from raft.tasks import task
from ..base.utils import notice, notice_end
from .base import AzureTask


@task(klass=AzureTask)
def apps(ctx, name=None, creds=None, quiet=False, **kwargs):
    """
    shows all applications with a substring of name
    """
    from msgraph.graph_service_client import GraphServiceClient
    from msgraph.generated.applications.applications_request_builder import (
        ApplicationsRequestBuilder as A,
    )
    from msgraph.generated.models.application_collection_response import (
        ApplicationCollectionResponse as AppResponse,
    )
    from ..base.utils import print_table
    client = GraphServiceClient(credentials=creds)
    name = (name or '').lower()
    result = []

    async def yield_applications(skip=0):
        params = A.ApplicationsRequestBuilderGetQueryParameters(
            count=True,
            top=999,
        )
        if name:
            notice('building search')
            pieces = name.split(' ')
            search_query = [ f'"displayName:{piece}"' for piece in pieces ]
            params.search = ' AND '.join(search_query)
            notice_end(params.search)
        builder = A.ApplicationsRequestBuilderGetRequestConfiguration(
            query_parameters=params,
            headers={
                'ConsistencyLevel': 'eventual',
            }
        )
        notice('executing search')
        response: AppResponse = await client.applications.get(builder)
        notice_end(f'{response.odata_count}')
        for x in response.value:
            yield x
        if response.odata_next_link:
            async for x in yield_applications(skip + len(response.value)):
                yield x

    async def get():
        async for x in yield_applications():
            result.append(x)

    asyncio.run(get())
    if result and not quiet:
        header = [ 'app_id', 'name', ]
        rows = []
        for row in result:
            rows.append([
                row.id,
                row.display_name,
            ])
        rows.sort(key=lambda lx: lx[1].lower())
        print_table(header, rows)
        return None
    return result



import os.path

from raft.tasks import task

from ...base.utils import load_yaml
from ...base.utils import notice, notice_end, print_table, dump_yaml
from ...base.utils import get_context_value
from ..base import AwsTask
from ..base import yielder


@task(klass=AwsTask)
def nameservers(ctx, zone_name, session=None, **kwargs):
    """
    lists all of the name servers for a particular zone name
    """
    route53 = session.client('route53')
    g = yielder(route53, 'list_hosted_zones', session)
    header = [ 'zone', 'id', 'nameserver' ]
    rows = []
    for x in g:
        if zone_name.lower() in x['Name']:
            notice(x['Name'])
            hosted_zone = route53.get_hosted_zone(Id=x['Id'])
            if 'DelegationSet' in hosted_zone:
                ns = hosted_zone['DelegationSet']['NameServers']
                for nameserver in ns:
                    rows.append([ x['Name'], x['Id'], nameserver ])
            else:
                rows.append([ x['Name'], x['Id'], '' ])
            notice_end()
    print_table(header, rows)


def recordset_key(values):
    t = values['Type']
    if t in ('A', 'CNAME'):
        t = 'ACNAME'
    return f"{values['Name']}::{t}"


def update_template(filepath, data):
    from ...base.utils import yaml_serializer
    notice(f'updating {filepath}')
    rt = yaml_serializer(typ='rt')
    if os.path.exists(filepath):
        with open(filepath, 'r') as f:
            template_body = rt.load(filepath)
        resources = template_body['Resources']
        resource_map = {
            recordset_key(values['Properties']): x
            for x, values in resources.items()
            if values['Type'] == 'AWS::Route53::RecordSet'
        }
        for key, values in data.items():
            rkey = recordset_key(values['Properties'])
            existing_key = resource_map.get(rkey)
            if existing_key:
                props = values['Properties']
                resources[existing_key]['Properties']['Type'] = props['Type']
                resource_records = props.get('ResourceRecords')
                if resource_records:
                    resources[existing_key]['Properties']['ResourceRecords'] = resource_records
                else:
                    alias_target = props.get('AliasTarget')
                    resources[existing_key]['Properties'][
                        'AliasTarget'] = alias_target
            else:
                resources[key] = values
    else:
        template_body = dict(
            AWSTemplateFormatVersion='2010-09-09',
            Description='imported via upholstery',
            Resources=data,
        )
        notice_end('not found')
    with open(filepath, 'w') as f:
        rt.dump(template_body, f)
    notice_end()


@task(klass=AwsTask)
def export_hosted_zone(ctx, domain=None, zone_id=None, filename=None, session=None, prefix=None, **kwargs):
    """
    creates cloudformation yaml for all recordsets in a hostedzone
    """
    route53 = session.client('route53')
    if not zone_id:
        notice('finding zone id')
        zones = list(yielder(route53, 'list_hosted_zones', session))
        for x in zones:
            name = x['Name']
            if domain in { name, name[:-1] }:
                zone_id = x['Id']
                domain = name
                break
        if not zone_id:
            notice_end(False)
            return
        notice_end(zone_id)
    if not domain:
        notice('finding domain')
        zones = list(yielder(route53, 'list_hosted_zones', session))
        for x in zones:
            if x['Id'] == zone_id:
                domain = x['Name']
                break
        if not domain:
            notice_end(False)
            return
        notice_end(domain)
    records = list(yielder(
        route53,
        'list_resource_record_sets',
        session,
        HostedZoneId=zone_id))
    resources = {}
    for x in records:
        name = x['Name']
        t = x['Type']
        properties = {
            'HostedZoneName': domain,
            'Name': name,
            'Type': t,
        }
        for key in 'TTL', 'ResourceRecords', 'AliasTarget':
            value = x.get(key)
            if value:
                if key == 'ResourceRecords':
                    properties[key] = [ y['Value'] for y in value ]
                else:
                    properties[key] = value
        name = name.rsplit(domain, 1)
        name = name[0][:-1].split('.')
        name = [ (piece if name != '.' else 'dot') for piece in name ]
        name = [ piece.title() for piece in name ]
        name = ''.join(name)
        name = name.replace('_', '')
        name = name or 'Apex'
        if t not in ('A', 'CNAME'):
            name = f'{name}{t.lower().title()}'
        name = name.replace('-', '')
        resources[name] = {
            'Type': 'AWS::Route53::RecordSet',
            'Properties': properties,
        }
    if filename:
        prefix = prefix or get_context_value(ctx, 'upholstery.prefix')
        _, ext = os.path.splitext(filename)
        ext = ext.lower()
        if ext not in ('.yml', '.yaml'):
            filename = f'{filename}.yml'
        filename = os.path.join(prefix, 'cfn', filename)
        update_template(filename, resources)
    else:
        dump_yaml(resources, quiet=False)



import contextlib
import json
import os
import signal
import sys
from subprocess import check_call

from raft.tasks import task

from convocations.base.utils import dump_yaml
from convocations.base.utils import get_context_value
from convocations.base.utils import notice, notice_end, print_table
from convocations.aws.base import AwsTask, yielder


def ssm_client(session, quiet=False):
    if not quiet:
        notice('connecting to ssm')
    ssm = session.client('ssm')
    if not quiet:
        notice_end()
    return ssm


def find_managed_instances(name, session, ssm=None, quiet=False):
    name = name.lower()
    ssm = ssm or ssm_client(session, quiet=quiet)
    if not quiet:
        notice('finding matching instances')
    rg = yielder(ssm, 'describe_instance_information', session)
    result = []
    for x in rg:
        st = x['ComputerName']
        instance_id = x['InstanceId']
        if name in st.lower() or name == instance_id:
            result.append(x)
    if not quiet:
        notice_end()
    return result


def instance_table(instances):
    header = [ 'id', 'name', 'ip' ]
    rows = []
    for x in instances:
        st = x['ComputerName']
        instance_id = x['InstanceId']
        row = [ instance_id, st, x['IPAddress'] ]
        rows.append(row)
    rows.sort(key=lambda lx: lx[1])
    print()
    print_table(header, rows)


@task(klass=AwsTask)
def managed_instances(ctx, name='', session=None, **kwargs):
    """
    lists all ssm managed instances with `name` in their `ComputerName`
    """
    rg = find_managed_instances(name, session)
    instance_table(rg)


@task(klass=AwsTask)
def unmanaged_instances(ctx, session=None, **kwargs):
    """
    lists all ec2 instances not enrolled in ssm management
    """
    from convocations.aws.base import name_tag
    from convocations.aws.ec2 import yield_instances
    rg = find_managed_instances(name='', session=session)
    managed_ids = { x['InstanceId'] for x in rg }
    notice('connecting to ec2')
    instances = yield_instances(session=session)
    instances = [ x for x in instances if x['InstanceId'] not in managed_ids ]
    notice_end()
    header = [ 'id', 'name', 'ip' ]
    rows = []
    for x in instances:
        row = [ x['InstanceId'], name_tag(x), x.get('PrivateIpAddress', '') ]
        rows.append(row)
    rows.sort(key=lambda lx: lx[1])
    print()
    print_table(header, rows)


@task(klass=AwsTask)
def ssm_documents(ctx, quiet=False, session=None, **kwargs):
    """
    prints a list of available ssm documents
    """
    ssm = ssm_client(session, quiet=quiet)
    filters = [{
        'Key': 'DocumentType',
        'Values': [ 'Session' ],
    }, {
        'Key': 'Owner',
        'Values': [ 'Amazon', ],
    }, {
        'Key': 'Name',
        'Values': [ 'SSM-SessionManager', ]
    }]
    rg = yielder(ssm, 'list_documents', session=session,
                 Filters=filters)
    header = [ 'name' ]
    rows = []
    for x in rg:
        rows.append([ x['Name'] ])
    print_table(header, rows)


@contextlib.contextmanager
def ignore_user_entered_signals():
    """
    Ignores user entered signals to avoid process getting killed.
    """
    if sys.platform == 'win32':
        signal_list = [signal.SIGINT]
    else:
        signal_list = [signal.SIGINT, signal.SIGQUIT, signal.SIGTSTP]
    actual_signals = []
    for user_signal in signal_list:
        actual_signals.append(signal.signal(user_signal, signal.SIG_IGN))
    try:
        yield
    finally:
        for sig, user_signal in enumerate(signal_list):
            signal.signal(user_signal, actual_signals[sig])


@task(klass=AwsTask, help={
    'name': 'the name or instance id of the target instance in ssm',
    'reason': 'the reason we are using ssm to connect, for auditing',
    'session': 'do not pass anything into this parameter',
})
def start_shell(ctx, name, reason='convocations', session=None, quiet=False, **kwargs):
    """
    starts a shell to the specified instance
    """
    ssm = ssm_client(session, quiet=quiet)
    rg = find_managed_instances(name, session, ssm=ssm, quiet=quiet)
    if not rg:
        notice_end('no matching instances found')
        return
    if len(rg) > 1:
        notice_end('2 or more matching instances found')
        instance_table(rg)
        return
    notice('checking plugin version')
    result = ctx.run('session-manager-plugin --version', hide=True, warn=False)
    if result.exited != 0:
        notice_end(False)
        notice_end('please install the session manager plugin')
        base_url = 'https://docs.aws.amazon.com/systems-manager/latest/userguide/'
        print(f'ubuntu: {base_url}install-plugin-debian-and-ubuntu.html')
        print(f'windows: {base_url}install-plugin-windows.html')
        return
    notice_end(f'{result.stdout}')
    instance_id = rg[0]['InstanceId']
    params = { 'Reason': reason, 'Target': instance_id, }
    start_session_result = ssm.start_session(**params)
    dump_yaml(start_session_result, quiet=False)
    profile = get_context_value(ctx, 'aws.profile')
    endpoint_url = ssm.meta.endpoint_url
    env = os.environ.copy()
    env_key = 'AWS_SSM_START_SESSION_RESPONSE'
    env[env_key] = json.dumps(start_session_result)
    with ignore_user_entered_signals():
        check_call([
            "session-manager-plugin",
            env_key,
            ssm.meta.region_name,
            "StartSession",
            profile,
            json.dumps(params),
            endpoint_url], env=env)



import asyncio
from raft.tasks import task
from .base import AzureTask
from ..base.utils import notice, notice_end, print_table


@task(klass=AzureTask)
def groups(ctx, name, creds=None, quiet=False, members=False, **kwargs):
    """
    shows all groups with a substring of name
    """
    from msgraph.graph_service_client import GraphServiceClient
    from msgraph.generated.models.group_collection_response import (
        GroupCollectionResponse
    )
    from msgraph.generated.groups.groups_request_builder import (
        GroupsRequestBuilder as G,
    )
    name = (name or '').lower()
    client = GraphServiceClient(credentials=creds)
    name = (name or '').lower()
    result = []

    async def yield_groups(skip=0):
        params = G.GroupsRequestBuilderGetQueryParameters(
            count=True,
            top=999,
        )
        if name:
            notice('building search')
            pieces = name.split(' ')
            search_query = [ f'"displayName:{piece}"' for piece in pieces ]
            params.search = ' AND '.join(search_query)
            notice_end(params.search)
        builder = G.GroupsRequestBuilderGetRequestConfiguration(
            query_parameters=params,
            headers={
                'ConsistencyLevel': 'eventual',
            }
        )
        notice('executing search')
        response: GroupCollectionResponse = await client.groups.get(builder)
        notice_end(f'{response.odata_count}')
        for x in response.value:
            yield x
        if response.odata_next_link:
            async for x in yield_groups(skip + len(response.value)):
                yield x

    async def get_members(g):
        from msgraph.generated.groups.item.members.members_request_builder import MembersRequestBuilder as M
        params = M.MembersRequestBuilderGetQueryParameters(
            select=[ 'id', 'userPrincipalName' ],
        )
        builder = M.MembersRequestBuilderGetRequestConfiguration(
            query_parameters=params,
        )
        response = await client.groups.by_group_id(g.id).members.get(builder)
        return response

    async def get():
        async for x in yield_groups():
            result.append(x)
            if members:
                notice(f'members / {x.display_name}')
                member_response = await get_members(x)
                notice_end()
                x.members = member_response.value

    asyncio.run(get())
    if result and not quiet:
        header = [ 'group_id', 'name', 'type', ]
        if members:
            header.append('member_id')
            header.append('user_principal_name')
        rows = []
        for row in result:
            types = []
            if row.mail_enabled:
                types.append('mail')
            if row.security_enabled:
                types.append('security')
            if not members:
                rows.append([
                    row.id,
                    row.display_name,
                    '; '.join(types),
                ])
            elif not row.members:
                rows.append([
                    row.id,
                    row.display_name,
                    '; '.join(types),
                    '',
                    '',
                ])
            else:
                for member in row.members:
                    rows.append([
                        row.id,
                        row.display_name,
                        '; '.join(types),
                        member.id,
                        member.user_principal_name,
                    ])
        if not members:
            rows.sort(key=lambda lx: lx[1].lower())
        else:
            rows.sort(key=lambda lx: (lx[1].lower(), lx[4].lower()))
        print_table(header, rows)
    else:
        return result


